"""
# Fearless lossy compression with `compression-safeguards`

Lossy compression can be *scary* as valuable information or features of the
data may be lost.

By using safeguards to **guarantee** your safety requirements, lossy
compression can be applied safely and *without fear*.

## Overview

This package provides several
[`Safeguard`][.safeguards.abc.Safeguard]s (refer to the
[`SafeguardKind`][.safeguards.SafeguardKind] for an enumeration) with which you
can express *your* requirements for lossy compression to be safe to use.

The safeguards are then combined in the
[`Safeguards`][.api.Safeguards], which can be used to compute and apply the
required correction to lossy-compressed data so that it satisfies your safety
guarantees.

This package provides the implementations of the safeguards and the low-level
`Safeguards` API. Please also refer to the following integrations of the
safeguards with popular compression APIs:

- [`numcodecs-safeguards`][numcodecs_safeguards]: provides the
  [`SafeguardedCodec`][numcodecs_safeguards.SafeguardedCodec] meta-compressor that
  conveniently applies safeguards to any compressor using the
  [`numcodecs.abc.Codec`][numcodecs.abc.Codec] API.
- [`xarray-safeguards`][xarray_safeguards]: provides functionality to use
  safeguards with (chunked) [`xarray.DataArray`][xarray.DataArray]s.

## Examples

In the following examples, we assume that there is *some* compressor that can
`compress` and `decompress` n-dimensional `data` using functions of that name.

### Basic Usage

You can guarantee an absolute error bound of $eb_{abs} = 0.1$  as follows:

<!--
```py
import numpy as np
def compress(data):
    return data
def decompress(data):
    return np.zeros_like(data)
```
-->
<!--pytest-codeblocks:cont-->
```py
import numpy as np
from compression_safeguards import Safeguards

# create the `Safeguards`
sg = Safeguards(safeguards=[
    # guarantee an absolute error bound of 0.1:
    #   |x - x'| <= 0.1
    dict(kind="eb", type="abs", eb=0.1),
])

# generate some random data to compress
data = np.random.normal(size=(10, 10, 10))

## compression

# compress and decompress the data using *some* compressor
compressed = compress(data)
decompressed = decompress(compressed)

# compute the correction that the safeguards would need to apply to
# guarantee the selected safety requirements
correction = sg.compute_correction(data, decompressed)

# now the compressed data and correction can be stored somewhere
# ...
# and loaded again to decompress

## decompression
decompressed = decompress(compressed)
decompressed = sg.apply_correction(decompressed, correction)

# the safeguard properties are now guaranteed to hold
assert np.all(np.abs(data - decompressed) <= 0.1)
```

### Instantiating the safeguards

The safeguards can be instantiated from JSON-like configuration:

```py
from compression_safeguards import Safeguards

sg = Safeguards(safeguards=[
    dict(kind="eb", type="abs", eb=0.1),
])
```

or by using the [`SafeguardKind`][.safeguards.SafeguardKind]:

```py
from compression_safeguards import Safeguards, SafeguardKind

sg = Safeguards(safeguards=[
    SafeguardKind.eb.value(type="abs", eb=0.1),
])
```

These two methods can be freely combined.

The entire safeguards can also be turned into a JSON configuration and
recreated from such configuration:

```py
from compression_safeguards import Safeguards

sg = Safeguards(safeguards=[
    dict(kind="eb", type="abs", eb=0.1),
])
config = sg.get_config()

sg = Safeguards.from_config(config)
assert sg.get_config() == config
```

### Combining several safeguards

All of the provided safeguards can be freely combined and are guaranteed to
work together.

Providing several safeguards means that *all* the specified
safety requirements must be upheld:

```py
from compression_safeguards import Safeguards

sg = Safeguards(safeguards=[
    # guarantee an absolute error bound
    dict(kind="eb", type="abs", eb=0.1),
    # and that the data sign is preserved
    dict(kind="sign"),
])
```

This package also provides several combinators that can be used to express
pointwise logical combinations of safeguards:

```py
from compression_safeguards import Safeguards

sg = Safeguards(safeguards=[
    # guarantee that, for each element, *both* an absolute error bound of 0.1
    # *and* a relative error bound of 1% are upheld
    dict(kind="all", safeguards=[
        dict(kind="eb", type="abs", eb=0.1),
        dict(kind="eb", type="rel", eb=0.01),
    ]),
])

sg = Safeguards(safeguards=[
    # guarantee that, for each element, an absolute error bound of 0.1
    # *or* a relative error bound of 1% are upheld
    dict(kind="any", safeguards=[
        dict(kind="eb", type="abs", eb=0.1),
        dict(kind="eb", type="rel", eb=0.01),
    ]),
])
```

### Regionally varying safeguards using late-bound parameters

By default, all safeguards apply the same safety guarantees across the entire
data domain. This package supports two approaches for regionally varying the
guarantees, i.e. applying different guarantees to different data regions.

First, the `select` combinator can be used to switch between two or more
safeguards (or safeguard combinations) using a selection indices array. Unlike
normal safeguard parameters, this selector is a late-bound parameter whose
value is not specified during safeguard initialisation but only later when the
safeguard is applied:

<!--
```py
import numpy as np
def compress(data):
    return data
def decompress(data):
    return np.zeros_like(data)
```
-->
<!--pytest-codeblocks:cont-->
```py
from compression_safeguards import Safeguards
from compression_safeguards.utils.bindings import Bindings

sg = Safeguards(safeguards=[
    # select between a coarser, medium, and finer absolute error bound
    #  safeguard based on the late-bound "mask" parameter
    dict(kind="select", selector="mask", safeguards=[
        dict(kind="eb", type="abs", eb=1.0),
        dict(kind="eb", type="abs", eb=0.1),
        dict(kind="eb", type="abs", eb=0.01),
    ]),
])

# generate some random data to compress
data = np.random.normal(size=(10, 10, 10))

## compression (now with late-bound parameters)

# compress and decompress the data using *some* compressor
compressed = compress(data)
decompressed = decompress(compressed)

# compute the correction that the safeguards would need to apply to
# guarantee the selected safety requirements
correction = sg.compute_correction(data, decompressed, late_bound=Bindings(
    # bind the selection mask that selects between the three safeguards
    # the mask must be broadcastable to the data shape
    mask=np.array([0, 0, 0, 1, 1, 1, 1, 2, 2, 2]).reshape((1, 10, 1)),
))

## decompression (as before)
decompressed = decompress(compressed)
decompressed = sg.apply_correction(decompressed, correction)
```

It is worth noting that the late-bound parameters are only needed at
compression time, decompression is unchanged.

While this first method can combine over any safeguards, it is only convenient
for selecting between a small number of different safeguards.

For the error bound safeguards (pointwise and on quantities of interest), the
error bounds themselves can be provided as late-bound parameters to allow for
smoothly varying error bounds across the data domain. The above example could
then be equivalently expressed as:

<!--
```py
import numpy as np
def compress(data):
    return data
def decompress(data):
    return np.zeros_like(data)
```
-->
<!--pytest-codeblocks:cont-->
```py
from compression_safeguards import Safeguards
from compression_safeguards.utils.bindings import Bindings

sg = Safeguards(safeguards=[
    # absolute error bound with a late-bound "eb" parameter
    dict(kind="eb", type="abs", eb="eb"),
])

# generate some random data to compress
data = np.random.normal(size=(10, 10, 10))

## compression (now with late-bound parameters)

# compress and decompress the data using *some* compressor
compressed = compress(data)
decompressed = decompress(compressed)

# compute the correction that the safeguards would need to apply to
# guarantee the selected safety requirements
correction = sg.compute_correction(data, decompressed, late_bound=Bindings(
    # bind the late-bound absolute error bound
    # the bound must be broadcastable to the data shape
    eb=np.array([
        1.0, 1.0, 1.0, 0.1, 0.1, 0.1, 0.1, 0.01, 0.01, 0.01,
    ]).reshape((1, 10, 1)),
))

## decompression (as before)
decompressed = decompress(compressed)
decompressed = sg.apply_correction(decompressed, correction)
```
"""

__all__ = ["Safeguards", "SafeguardKind"]

from .api import Safeguards
from .safeguards import SafeguardKind



import itertools
from collections.abc import Callable, Iterator, Mapping
from typing import Self

import numpy as np
from typing_extensions import override  # MSPV 3.12

from ....utils.bindings import Parameter
from ....utils.error import ctx
from ..typing import Es, F, Ns, Ps, np_sndarray
from .abc import AnyExpr, Expr
from .addsub import ScalarLeftAssociativeSum
from .constfold import ScalarFoldedConstant
from .data import Data
from .divmul import ScalarMultiply
from .group import Group


class Array(Expr[AnyExpr, *tuple[AnyExpr, ...]]):
    __slots__: tuple[str, ...] = ("_array",)
    _array: np.ndarray

    def __init__(self, el: AnyExpr, *els: AnyExpr) -> None:
        if isinstance(el, Array):
            aels = [el._array]
            for e in els:
                if not (isinstance(e, Array) and e.shape == el.shape):
                    raise (
                        ValueError(
                            "elements must all have the consistent shape "
                            + f"{el.shape}"
                        )
                        | ctx
                    )
                aels.append(e._array)
            self._array = np.array(aels, copy=None)
        else:
            for e in els:
                if isinstance(e, Array):
                    raise ValueError("elements must all be scalar") | ctx
            self._array = np.array((el, *els), copy=None)

    @staticmethod
    def from_data_shape(shape: tuple[int, ...]) -> "Array":
        out = Array.__new__(Array)
        out._array = np.empty(shape, dtype=object)

        for i in itertools.product(*[range(a) for a in shape]):
            out._array[i] = Data(index=i)

        return out

    @property
    @override
    def args(self) -> tuple[AnyExpr, *tuple[AnyExpr, ...]]:
        if self._array.ndim == 1:
            return tuple(self._array)
        return tuple(Array(*a) for a in self._array)  # type: ignore

    @override
    def with_args(self, el: AnyExpr, *els: AnyExpr) -> Self:
        return type(self)(el, *els)

    @override
    def constant_fold(self, dtype: np.dtype[F]) -> F | AnyExpr:
        return Array.map(
            lambda e: ScalarFoldedConstant.constant_fold_expr(e, dtype),
            self,
        )

    @override
    def eval(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        assert False, "cannot evaluate an array expression"

    @override
    def compute_data_bounds_unchecked(
        self,
        expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        assert False, "cannot derive data bounds over an array expression"

    @property
    def shape(self) -> tuple[int, ...]:
        return self._array.shape

    @property
    def size(self) -> int:
        return self._array.size

    @staticmethod
    def map(map: Callable[[*Es], AnyExpr], *exprs: *Es) -> AnyExpr:
        if not any(isinstance(e, Array) for e in exprs):
            return map(*exprs)

        shape = None
        size = 0
        iters: list[Iterator[AnyExpr]] = []

        for i, expr in enumerate(exprs):  # type: ignore
            if isinstance(expr, Array):
                if shape is not None and expr.shape != shape:
                    raise (
                        ValueError(
                            f"shape mismatch between operands, expected {shape}"
                            + f" but found {expr.shape} for operand {i + 1}"
                        )
                        | ctx
                    )
                shape = expr.shape
                size = expr._array.size
                iters.append(expr._array.flat)
            else:
                iters.append(itertools.repeat(expr))

        out = Array.__new__(Array)
        out._array = np.fromiter(
            (map(*its) for its in zip(*iters)),
            dtype=object,
            count=size,
        ).reshape(shape)
        return out

    def index(self, index: tuple[int | slice, ...]) -> AnyExpr:
        a = self._array[index]
        if isinstance(a, np.ndarray):
            out = Array.__new__(Array)
            out._array = a
            return out
        return a

    def transpose(self) -> Self:
        out = Array.__new__(type(self))
        out._array = self._array.T
        return out

    def sum(self) -> AnyExpr:
        sum_: AnyExpr = ScalarLeftAssociativeSum(*list(self._array.flat))
        # we can return a group here since sum_ is not an array
        assert not isinstance(sum_, Array)
        return Group(sum_)

    def flatlist(self) -> list[AnyExpr]:
        return list(self._array.flat)

    @staticmethod
    def matmul(
        left: "Array",
        right: "Array",
    ) -> "Array":
        if len(left.shape) != 2:
            raise ValueError("can only matmul(a, b) a 2D array a") | ctx
        if len(right.shape) != 2:
            raise ValueError("can only matmul(a, b) a 2D array b") | ctx
        if left.shape[1] != right.shape[0]:
            raise (
                ValueError(
                    "can only matmul(a, b) with shapes (n, k) x (k, m) -> "
                    + f"(n, m) but got {left.shape} x {right.shape}"
                )
                | ctx
            )
        out = Array.__new__(Array)
        out._array = np.empty((left.shape[0], right.shape[1]), dtype=object)
        for n in range(left.shape[0]):
            for m in range(right.shape[1]):
                sum_: AnyExpr = ScalarLeftAssociativeSum(
                    *[
                        ScalarMultiply(left._array[n, k], right._array[k, m])
                        for k in range(left.shape[1])
                    ]
                )
                # we can apply a group here since sum_ is not an array
                assert not isinstance(sum_, Array)
                out._array[n, m] = Group(sum_)
        return out

    @override
    def __repr__(self) -> str:
        return f"{self._array!r}".removeprefix("array(").removesuffix(", dtype=object)")



from traceback import print_exception

import numpy as np

from ..qois import StencilQuantityOfInterestExpression
from . import StencilQuantityOfInterest
from .expr.constfold import ScalarFoldedConstant

if __name__ == "__main__":
    while True:
        try:
            qoi = input("qoi > ")
        except EOFError:
            break
        if len(qoi) == 0:
            break
        try:
            qoi_expr = StencilQuantityOfInterest(
                StencilQuantityOfInterestExpression(qoi),
                stencil_shape=(3,),
                stencil_I=(1,),
            )
            print(f"parsed: {qoi_expr!r}")  # noqa: T201
            Xs: np.ndarray[tuple[int, int], np.dtype[np.float64]] = (
                np.arange(3).astype(np.float64).reshape(1, 3)
            )
            print(  # noqa: T201
                f"folded: {ScalarFoldedConstant.constant_fold_expr(qoi_expr._expr, Xs.dtype)!r}"
            )
            print(f"eval: {qoi_expr.eval(Xs, dict())}")  # noqa: T201
        except Exception as err:
            print_exception(err, limit=0)



import itertools
from collections.abc import Mapping
from contextlib import contextmanager
from hashlib import blake2b
from typing import Literal

import numpy as np
from typing_extensions import override  # MSPV 3.12

from ....utils._compat import _broadcast_to, _ensure_array, _ones, _zeros
from ....utils.bindings import Bindings, Parameter
from ....utils.cast import from_float
from ....utils.error import ctx
from ....utils.intervals import Interval, IntervalUnion
from ....utils.typing import S, T
from ..bound import DataBounds, data_bounds
from ..typing import F, Ns, Ps, np_sndarray
from .abc import AnyExpr, EmptyExpr


class HashingExpr(EmptyExpr):
    __slots__: tuple[str, ...] = ("_data_indices", "_late_bound_constants")
    _data_indices: frozenset[tuple[int, ...]]
    _late_bound_constants: frozenset[Parameter]

    def __init__(
        self,
        data_indices: frozenset[tuple[int, ...]],
        late_bound_constants: frozenset[Parameter],
    ) -> None:
        self._data_indices = data_indices
        self._late_bound_constants = late_bound_constants

    @staticmethod
    def from_data_shape(
        data_shape: tuple[int, ...], late_bound_constants: frozenset[Parameter]
    ) -> "HashingExpr":
        data_indices = frozenset(
            [tuple(i) for i in itertools.product(*[range(a) for a in data_shape])]
        )

        return HashingExpr(data_indices, late_bound_constants)

    @property
    @override
    def args(self) -> tuple[()]:
        return ()

    @override
    def with_args(self) -> "HashingExpr":
        return HashingExpr(self._data_indices, self._late_bound_constants)

    @property  # type: ignore
    @override
    def has_data(self) -> bool:
        return True

    @override  # type: ignore
    def eval_has_data(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[np.bool]]:
        return _ones(Xs.shape[:1], dtype=np.dtype(np.bool))

    @property  # type: ignore
    @override
    def data_indices(self) -> frozenset[tuple[int, ...]]:
        return self._data_indices

    @override  # type: ignore
    def apply_array_element_offset(
        self,
        axis: int,
        offset: int,
    ) -> "HashingExpr":
        data_indices = []
        for index_ in self._data_indices:
            index = list(index_)
            index[axis] += offset
            data_indices.append(tuple(index))
        return HashingExpr(
            data_indices=frozenset(data_indices),
            late_bound_constants=self._late_bound_constants,
        )

    @property  # type: ignore
    @override
    def late_bound_constants(self) -> frozenset[Parameter]:
        return self._late_bound_constants

    @override
    def constant_fold(self, dtype: np.dtype[F]) -> F | AnyExpr:
        return self

    @override
    def eval(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        x_size: Ps = Xs.shape[0]

        Xs_flat = Xs.reshape((x_size, Xs.size // x_size))
        late_bound_flat = {
            param: value.reshape((x_size, Xs.size // x_size))
            for param, value in late_bound.items()
        }

        hash: np.ndarray[tuple[Ps], np.dtype[F]] = np.empty(x_size, dtype=Xs.dtype)

        # hash Xs and late_bound stencils for every element in X
        for i in range(x_size):
            hasher = blake2b(digest_size=Xs.dtype.itemsize)
            hasher.update(Xs_flat[i].tobytes())
            for param, value_flat in sorted(
                late_bound_flat.items(), key=lambda pv: pv[0]
            ):
                hasher.update(param.encode())
                hasher.update(value_flat[i].tobytes())
            hash[i] = np.frombuffer(hasher.digest(), dtype=Xs.dtype, count=1)[0]

        return hash

    @data_bounds(DataBounds.infallible)
    @override
    def compute_data_bounds_unchecked(
        self,
        expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        X_hash: np.ndarray[tuple[Ps], np.dtype[F]] = self.eval(Xs, late_bound)
        Xs_hash: np_sndarray[Ps, Ns, np.dtype[F]] = _broadcast_to(
            X_hash.reshape(Xs.shape[:1] + (1,) * (Xs.ndim - 1)), Xs.shape
        )
        return _ensure_array(Xs_hash, copy=True), _ensure_array(Xs_hash, copy=True)

    @override
    def __repr__(self) -> str:
        return "hashing"


@contextmanager
def _patch_for_hashing_qoi_dev_only():
    from ...stencil.qoi import eb  # noqa: PLC0415

    old_interval_union_contains = IntervalUnion.contains
    old_interval_union_pick = IntervalUnion.pick

    old_stencil_qoi_check_pointwise = (
        eb.StencilQuantityOfInterestErrorBoundSafeguard.check_pointwise
    )
    old_compute_safe_data_lower_upper_interval_union = (
        eb.compute_safe_data_lower_upper_interval_union
    )

    IntervalUnion.contains = _interval_union_contains
    IntervalUnion.pick = _interval_union_pick

    eb.StencilQuantityOfInterestErrorBoundSafeguard.check_pointwise = (
        _stencil_qoi_check_pointwise
    )
    eb.compute_safe_data_lower_upper_interval_union = (
        _compute_safe_data_lower_upper_interval_union
    )

    try:
        yield
    finally:
        IntervalUnion.contains = old_interval_union_contains
        IntervalUnion.pick = old_interval_union_pick

        eb.StencilQuantityOfInterestErrorBoundSafeguard.check_pointwise = (
            old_stencil_qoi_check_pointwise
        )
        eb.compute_safe_data_lower_upper_interval_union = (
            old_compute_safe_data_lower_upper_interval_union
        )


def _interval_union_contains(
    self, other: np.ndarray[S, np.dtype[T]]
) -> np.ndarray[S, np.dtype[np.bool]]:
    return _ones(other.shape, dtype=np.dtype(np.bool))


def _interval_union_pick(
    self, prediction: np.ndarray[S, np.dtype[T]]
) -> np.ndarray[S, np.dtype[T]]:
    if np.all(self._lower == 1):
        raise ValueError("fuzzer hash is all ones") | ctx
    return _ensure_array(self._lower[0].reshape(prediction.shape), copy=True)


def _stencil_qoi_check_pointwise(
    self,
    data: np.ndarray[S, np.dtype[T]],
    prediction: np.ndarray[S, np.dtype[T]],
    *,
    late_bound: Bindings,
    where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
) -> np.ndarray[S, np.dtype[np.bool]]:
    if np.all(prediction == 1):
        return _zeros(data.shape, dtype=np.dtype(np.bool))
    return _ones(data.shape, dtype=np.dtype(np.bool))


def _compute_safe_data_lower_upper_interval_union(
    data: np.ndarray[S, np.dtype[T]],
    data_float_lower: np.ndarray[S, np.dtype[F]],
    data_float_upper: np.ndarray[S, np.dtype[F]],
) -> IntervalUnion[T, int, int]:
    valid = Interval.empty_like(data)
    valid._lower[:] = from_float(data_float_lower, data.dtype).flatten()
    valid._upper[:] = from_float(data_float_lower, data.dtype).flatten()
    return valid.into_union()



from collections.abc import Mapping

import numpy as np
from typing_extensions import override  # MSPV 3.12

from ....utils._compat import _is_positive_zero
from ....utils.bindings import Parameter
from ..bound import checked_data_bounds
from ..typing import F, Ns, Ps, np_sndarray
from .abc import AnyExpr, Expr
from .constfold import ScalarFoldedConstant


class ScalarSign(Expr[AnyExpr]):
    __slots__: tuple[str, ...] = ("_a",)
    _a: AnyExpr

    def __init__(self, a: AnyExpr):
        self._a = a

    @property
    @override
    def args(self) -> tuple[AnyExpr]:
        return (self._a,)

    @override
    def with_args(self, a: AnyExpr) -> "ScalarSign":
        return ScalarSign(a)

    @override
    def constant_fold(self, dtype: np.dtype[F]) -> F | AnyExpr:
        return ScalarFoldedConstant.constant_fold_unary(
            self._a,
            dtype,
            np.sign,
            ScalarSign,
        )

    @override
    def eval(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        return np.sign(self._a.eval(Xs, late_bound))

    @checked_data_bounds
    @override
    def compute_data_bounds_unchecked(
        self,
        expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        # evaluate arg and sign(arg)
        arg = self._a
        argv = arg.eval(Xs, late_bound)
        exprv: np.ndarray[tuple[Ps], np.dtype[F]] = np.sign(argv)

        # evaluate the lower and upper sign bounds that satisfy the expression bound
        expr_lower = np.ceil(expr_lower)
        expr_upper = np.floor(expr_upper)

        smallest_subnormal = np.finfo(Xs.dtype).smallest_subnormal

        # compute the lower and upper arg bounds that produce the sign bounds
        # sign(-0.0) = +0.0 and sign(+0.0) = +0.0
        # - expr_lower = -0.0 (exprv >= +0.0) -> arg_lower = -0.0, exprv = +0.0
        # - expr_lower = +0.0 (exprv >= +0.0) -> arg_lower = -0.0, exprv = +0.0
        # - expr_upper = -0.0 (exprv < 0) -> arg_upper = -subnormal, exprv = -1
        # - expr_upper = +0.0 (exprv <= +0.0) -> arg_upper = +0.0, exprv = +0.0
        arg_lower: np.ndarray[tuple[Ps], np.dtype[F]] = np.full(
            Xs.shape[:1], smallest_subnormal
        )
        arg_lower[np.less_equal(expr_lower, -1)] = -np.inf
        arg_lower[expr_lower == 0] = -0.0
        np.copyto(arg_lower, exprv, where=np.isnan(exprv), casting="no")

        arg_upper: np.ndarray[tuple[Ps], np.dtype[F]] = np.full(
            Xs.shape[:1], -smallest_subnormal
        )
        arg_upper[np.greater_equal(expr_upper, +1)] = np.inf
        arg_upper[_is_positive_zero(expr_upper)] = +0.0
        np.copyto(arg_upper, exprv, where=np.isnan(exprv), casting="no")

        return arg.compute_data_bounds(
            arg_lower,
            arg_upper,
            Xs,
            late_bound,
        )

    @override
    def __repr__(self) -> str:
        return f"sign({self._a!r})"
